import ij.*;
import ij.plugin.filter.PlugInFilter;
import ij.process.*;
import ij.gui.*;

//Binary Filter Reconstruct by Gabriel Landini, G.Landini@bham.ac.uk
// v1.2  20/11/2004  ~x4 speed increase in the reconstruction
// v1.3 29?Jun/2006 speed up via floodfilling instead of old method of reconstruction

public class BinaryFilterReconstruct_ implements PlugInFilter {
	protected int nerosions;
	protected boolean doIwhite;

	public int setup(String arg, ImagePlus imp) {
		ImageStatistics stats;
		stats=imp.getStatistics();
		if (stats.histogram[0]+stats.histogram[255]!=stats.pixelCount){
			IJ.error("8-bit binary image (0 and 255) required.");
			return DONE;
		}

		if (arg.equals("about"))
			{showAbout(); return DONE;}
		GenericDialog gd = new GenericDialog("Binary Filter Reconstruct", IJ.getInstance());
		gd.addMessage("Binary Filter Reconstruct v1.3");
		gd.addNumericField("Erosions",1,0);
		gd.addCheckbox("White particles on black background",true);

		gd.showDialog();
		if (gd.wasCanceled())
			return DONE;
        nerosions=(int) gd.getNextNumber();
		doIwhite = gd.getNextBoolean ();
		return DOES_8G+DOES_STACKS;
	}

	public void run(ImageProcessor ip) {
		int width = ip.getWidth();
		int height = ip.getHeight();
		int x, y, i;
		int [][] mask = new int [width][height];
		int [][] seed = new int [width][height];
		//get original

		if (!doIwhite){
			for(y=0;y<height;y++) {
				for(x=0;x<width;x++)
					ip.putPixel(x, y, 255-ip.getPixel(x,y));
			}

		}

		for(y=0;y<height;y++) {
			for(x=0;x<width;x++)
				mask[x][y]=ip.getPixel(x,y);
		}

		//erode n times
		for(i=0;i<nerosions;i++){
			if (doIwhite)
				ip.dilate(); // erodes black
			else
				ip.erode();

		}

		for(y=0;y<height;y++) {
			for(x=0;x<width;x++){
				seed[x][y]=ip.getPixel(x, y);
				if (mask[x][y]==255)
					ip.putPixel(x, y, 127); //put mask value
				else
					ip.putPixel(x, y, 0); //put a 0
			}
		}

		FloodFiller ff = new FloodFiller(ip);

		ip.setColor(255);

		if (doIwhite){
			for(y=0;y<height;y++) {
				for(x=0;x<width;x++){
					if (seed[x][y]==255) {
						if (ip.getPixel(x, y)==127)
							ff.fill8(x, y);
					}
				}
			}
		}
		else{
			for(y=0;y<height;y++) {
				for(x=0;x<width;x++){
					if (seed[x][y]==0) {
						if (ip.getPixel(x, y)==127)
							ff.fill8(x, y);
					}
				}
			}
		}

		for(y=0;y<height;y++) {
			for(x=0;x<width;x++){
				if (ip.getPixel(x, y)==127)
					ip.putPixel(x, y, 0); //erase mask
			}
		}


		if (!doIwhite)
			ip.invert();
	}

	void showAbout() {
		IJ.showMessage("About BinaryFilterReconstruct_...",
		"BinaryFilterReconstruct_ by Gabriel Landini,  G.Landini@bham.ac.uk\n"+
		"ImageJ plugin for deleting particles that would disappear after\n"+
		"n erosions. The advantage over morphological opening is that\n"+
		"this plugin preserves the original shape of particles that\n"+
		"are not deleted (i.e. no particle smoothing takes place).\n"+
		"Supports both black and white binary particles.");
	}

}
