import ij.*;
import ij.plugin.*;
import ij.gui.*;
import ij.process.*;


/*
Binary reconstruction, by G.Landini. 22/Oct/2003 for ImageJ

Reconstructs binary objects of the "mask" image, based on markers present in the "seed" image.

It can handle black [0] particles on white [255] background or white particles
on a black background.

Binary reconstruction removes binary objects from the mask image which have
no markers in the seed image.

The reconstruction algorithm is: dilation (or erosion for white particles) of the
seed image AND mask image until there are no more pixles added by succesive
dilations (this is called the idempotent limit).

It supports stacks reconstructed by a single seed image and also
a single image reconstructed by a stack of seeds.
Not tested in versions older than 1.31l

Send any improvements or bugs to G.Landini@bham.ac.uk

This plugin is based on Calculator_Plus by Wayne Rasband (wayne@codon.nih.gov)

v1.0 release
v1.1 2/11/2003 replaced IJ erosion/dilations with own version as IJ does not process borders
v1.2 3/11/2003 small speed up
v1.3 16/4/2004 small speed up, feedback when reconstructing
v1.4 20/11/2004 ~x4 speedup with new dilation
v1.5 24/5/2006 changed 'dilations in a mask' for 'floodfill8 in the mask from the seed'
*/

public class BinaryReconstruct_ implements PlugIn {

	static String title = "Binary Reconstruct";
    static boolean createWindow = true, whiteparticles = true;
	int[] wList;
	private String[] titles;
	int i1Index;
	int i2Index;
	ImagePlus i1;
	ImagePlus i2;
	boolean replicate;

	public void run(String arg) {
		if (IJ.versionLessThan("1.37f"))
			return;
		wList = WindowManager.getIDList();
		if (wList==null || wList.length<2) {
			IJ.showMessage(title, "There must be at least two windows open");
			return;
		}
		titles = new String[wList.length];
		for (int i=0; i<wList.length; i++) {
			ImagePlus imp = WindowManager.getImage(wList[i]);
			if (imp!=null)
				titles[i] = imp.getTitle();
			else
				titles[i] = "";
		}

		if (!showDialog())
			return;

		long start = System.currentTimeMillis();
		boolean calibrated = i1.getCalibration().calibrated() || i2.getCalibration().calibrated();

		if (calibrated)
			createWindow = true;
		if (createWindow) {
			if (replicate)
				i2 = replicateImage(i2, calibrated, i1.getStackSize());
			else
				i2 = duplicateImage(i2, calibrated);
			if (i2==null)
				{IJ.showMessage(title, "Out of memory"); return;}
			i2.show();
		}
		reconstruct(i1, i2);
		IJ.showStatus(IJ.d2s((System.currentTimeMillis()-start)/1000.0, 2)+" seconds");
	}

	public boolean showDialog() {
		GenericDialog gd = new GenericDialog(title);
		gd.addMessage("Binary Reconstruction v 1.5");
		gd.addChoice("mask i1:", titles, titles[0]);
		gd.addChoice("seed i2:", titles, titles[1]);
		gd.addCheckbox("Create New Window", createWindow);
		gd.addCheckbox("White particles on black background", whiteparticles);
		gd.showDialog();
		if (gd.wasCanceled())
			return false;
		int i1Index = gd.getNextChoiceIndex();
		int i2Index = gd.getNextChoiceIndex();
		createWindow = gd.getNextBoolean();
		whiteparticles = gd.getNextBoolean();
		i1 = WindowManager.getImage(wList[i1Index]);
		i2 = WindowManager.getImage(wList[i2Index]);
		int d1 = i1.getStackSize();
		int d2 = i2.getStackSize();
		if (d2==1 && d1>1) {
			createWindow = true;
			replicate = true;
		}
		return true;
 }

	public void reconstruct(ImagePlus i1, ImagePlus i2) {
		int width  = i1.getWidth();
		int height = i1.getHeight();
		ImageProcessor ip1, ip2;
		int slices1 = i1.getStackSize();
		int slices2 = i2.getStackSize();
		float[] ctable1 = i1.getCalibration().getCTable();
		float[] ctable2 = i2.getCalibration().getCTable();
		ImageStack stack1 = i1.getStack();
		ImageStack stack2 = i2.getStack();
		int currentSlice = i2.getCurrentSlice();
		ImageStatistics stats;
		int x, y;


		//int [][] mask = new int [width][height];
		int [][] seed = new int [width][height];

		IJ.showStatus("Binary Reconstruction...");
		for (int n=1; n<=slices2; n++) {
			ip1 = stack1.getProcessor(n<=slices1?n:slices1);
			stats=i1.getStatistics();
			if (stats.histogram[0]+stats.histogram[255]!=stats.pixelCount){
				IJ.error("8-bit binary image (0 and 255) required.");
				return;
			}
			stats=i2.getStatistics();
				if (stats.histogram[0]+stats.histogram[255]!=stats.pixelCount){
				IJ.error("8-bit binary image (0 and 255) required.");
				return;
			}

			ip2 = stack2.getProcessor(n);
			ip1.setCalibrationTable(ctable1);
			ip2.setCalibrationTable(ctable2);

			if (!whiteparticles){
				ip1.invert();
				ip2.invert();
			}

			for(y=0;y<height;y++) {
				for(x=0;x<width;x++){
					seed[x][y]=ip2.getPixel(x, y);
					if (ip1.getPixel(x, y)==255)
						ip2.putPixel(x, y, 127); //put mask value
					else
						ip2.putPixel(x, y, 0); //put a 0
				}
			}
			i2.updateAndDraw();

			FloodFiller ff = new FloodFiller(ip2);

			ip2.setColor(255);
			for(y=0;y<height;y++) {
				for(x=0;x<width;x++){
					if (seed[x][y]==255) {
						if (ip2.getPixel(x, y)==127)
							ff.fill8(x, y);
					}
				}
			}

			for(y=0;y<height;y++) {
				for(x=0;x<width;x++){
					if (ip2.getPixel(x, y)==127)
						ip2.putPixel(x, y, 0); //erase mask
				}
			}



			i2.updateAndDraw();

			if (!whiteparticles){
				ip1.invert();
				ip2.invert();
			}

			if (n==currentSlice) {
				i2.getProcessor().resetMinAndMax();
				i2.updateAndDraw();
			}

			IJ.showProgress((double)n/slices2);
			IJ.showStatus(n+"/"+slices2);
		}
	}


	ImagePlus duplicateImage(ImagePlus img1, boolean calibrated) {
		ImageStack stack1 = img1.getStack();
		int n = stack1.getSize();
		ImageStack stack2 = img1.createEmptyStack();
		float[] ctable = img1.getCalibration().getCTable();
		try {
			for (int i=1; i<=n; i++) {
				ImageProcessor ip1 = stack1.getProcessor(i);
				ImageProcessor ip2 = ip1.duplicate();
				if (calibrated) {
					ip2.setCalibrationTable(ctable);
					ip2 = ip2.convertToFloat();
				}
				stack2.addSlice(stack1.getSliceLabel(i), ip2);
			}
		}
		catch(OutOfMemoryError e) {
			stack2.trim();
			stack2 = null;
			return null;
		}
		ImagePlus img2 =  new ImagePlus("Reconstructed", stack2);
		return img2;
	}

	ImagePlus replicateImage(ImagePlus img1, boolean calibrated, int n) {
		ImageProcessor ip1 = img1.getProcessor();
		ImageStack stack2 = img1.createEmptyStack();
		float[] ctable = img1.getCalibration().getCTable();
		try {
			for (int i=1; i<=n; i++) {
				ImageProcessor ip2 = ip1.duplicate();
				if (calibrated) {
					ip2.setCalibrationTable(ctable);
					ip2 = ip2.convertToFloat();
				}
				stack2.addSlice(null, ip2);
			}
		}
		catch(OutOfMemoryError e) {
			stack2.trim();
			stack2 = null;
			return null;
		}
		ImagePlus img2 =  new ImagePlus("Reconstructed", stack2);
		return img2;
	}

}

